from datetime import datetime, timedelta
from uuid import UUID

from fastapi import Depends, HTTPException
from fastapi.security import OAuth2PasswordBearer, OAuth2PasswordRequestForm
from jose import JWTError, jwt
from passlib.context import CryptContext

from ..schemas import TokenSchema, UserAuthDTO
from ..uow import UnitOfWork

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")

oauth2_schema = OAuth2PasswordBearer(tokenUrl="/token")


async def get_current_user(token: str = Depends(oauth2_schema)):
    if token is None:
        raise HTTPException(401, "Invalid credentials")

    try:
        payload = jwt.decode(token, "fsgddfsgdfgs", algorithms=["HS256"])

        id: UUID = UUID(payload.get("id"))
        name: str = payload.get("name", "")
        sub: str = payload.get("sub", "")
        expires_at: str = payload.get("expires_at", "")

        if not sub:
            raise HTTPException(401, "Invalid credentials")

        if expires_at:
            if is_expired(expires_at):
                raise HTTPException(401, "Invalid credentials")
        return UserAuthDTO(id=id, name=name, email=sub)
    except JWTError:
        raise HTTPException(401, "Invalid credentials")


def is_expired(expires_at: str) -> bool:
    """Return :obj:`True` if token has expired."""

    return datetime.strptime(expires_at, "%Y-%m-%d %H:%M:%S") < datetime.utcnow()


class AuthService:
    def __init__(self, uow: UnitOfWork) -> None:
        self.uow = uow
        self.crypto_context = CryptContext(schemes="bcrypt")

    async def authenticate(self, login: OAuth2PasswordRequestForm = Depends()) -> TokenSchema | None:
        async with self.uow:
            user = await self.uow.users.find_one(filter={"email": login.username})

        if user is None or user.hashed_password is None:
            raise HTTPException(401, "Incorrect password")
        else:
            if not self.crypto_context.verify(login.password, user.hashed_password):
                raise HTTPException(401, "Incorrect password")
            else:
                access_token = self._create_access_token(user.id, user.name, user.email)
                return TokenSchema(access_token=access_token, token_type="bearer")

    def _create_access_token(self, id: UUID, name: str, email: str) -> str:
        """Encode user information and expiration time."""

        payload = {
            "id": str(id),
            "name": name,
            "sub": email,
            "expires_at": self._expiration_time(),
        }

        print(payload)
        return jwt.encode(payload, "fsgddfsgdfgs", algorithm="HS256")

    @staticmethod
    def _expiration_time() -> str:
        """Get token expiration time."""

        expires_at = datetime.utcnow() + timedelta(minutes=30)
        return expires_at.strftime("%Y-%m-%d %H:%M:%S")
